#include "libkycpu.h"
#include <cstring-extension.h>
#include <errno.h>
#include <libkylog.h>
#include <sdkmarcos.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef __linux__
#include <sys/utsname.h>
#endif

struct _cpuInfo{
    size_t ncpus;               // CPU数量
    size_t corenums;            // CPU核心数量
    size_t wordorder;              // 字节序
    size_t cacheLevel;             // 缓存级别数
    size_t cpu_process;          // cpu线程数
    char *arch;                 // 架构
    char *vendor;               // 生产商
    char *model;                // 型号
    char *cur_freq_MHz;         // 主频
    char *min_freq_MHz;         // 最小主频
    char *max_freq_MHz;         // 最大主频
    char *flags;                // CPU 特性
    char *virt;                 // 虚拟化支持
};

struct _cpuInfo *cpuinf;

static void _free_cpuinfo()
{
    if (cpuinf)
    {
        SAFE_FREE(cpuinf->arch);
        SAFE_FREE(cpuinf->vendor);
        SAFE_FREE(cpuinf->model);
        SAFE_FREE(cpuinf->cur_freq_MHz);
        SAFE_FREE(cpuinf->max_freq_MHz);
        SAFE_FREE(cpuinf->min_freq_MHz);

        SAFE_FREE(cpuinf);
    }
}

static int lookup(char *line, char *pattern, char **value)
{
	char *p, *v;
	int len = strlen(pattern);
    // klog_debug("%s\n",pattern);
	/* don't re-fill already found tags, first one wins */
    if (strcmp(pattern,"processor")){
        if (!*line || *value)
		    return 0;
    }

	/* pattern */
	if (strncmp(line, pattern, len))
		return 0;

	/* white spaces */
	for (p = line + len; isspace(*p); p++);

	/* separator */
	if (*p != ':')
		return 0;

	/* white spaces */
	for (++p; isspace(*p); p++);

	/* value */
	if (!*p)
		return 0;
	v = p;

	/* end of value */
	len = strlen(line) - 1;
	for (p = line + len; isspace(*(p-1)); p--);
	*p = '\0';

	*value = strdup(v);
	return 1;
}

static void _get_cpu_info()
{
    // 我知道这里有竞态问题，但是不想引入pthread，所以算了
    if (!cpuinf)
    {
        cpuinf = (struct _cpuInfo *)calloc(1, sizeof(struct _cpuInfo));
        if (!cpuinf)
        {
            klog_err("内存分配失败：%s\n", strerror(errno));
            return ;
        }
    }
#ifdef __linux__
    struct utsname utsbuf;
    if (uname(&utsbuf) == -1)
    {
        klog_err("uname 失败：%s\n", strerror(errno));
        SAFE_FREE(cpuinf);
        return ;
    }

    cpuinf->arch = strdup(utsbuf.machine);

    FILE *fp = fopen("/proc/cpuinfo", "rt");
    if (!fp)
    {
        klog_err("/proc/cpuinfo 读取失败：%s\n", strerror(errno));
        SAFE_FREE(cpuinf);
        return ;
    }

#define CPUINF_BUFSIZE  8192
    char buffer[CPUINF_BUFSIZE];
    char *corenums;
    char *cpu_process;
    while (fgets(buffer, CPUINF_BUFSIZE, fp))
    {
        // klog_debug("%s",buffer);
        // if (strlen(strskipspace(buffer)) == 0)  // only '\n' left
        //     break;
        if (lookup(buffer, "vendor", &cpuinf->vendor));
        else if (lookup(buffer, "vendor_id", &cpuinf->vendor));
        else if (lookup(buffer, "model name", &cpuinf->model));
        else if (lookup(buffer, "flags", &cpuinf->flags));      // x86
        else if (lookup(buffer, "features", &cpuinf->flags));   // s390
        else if (lookup(buffer, "Features", &cpuinf->flags));   // aarch64
        else if (lookup(buffer, "type", &cpuinf->flags));       // sparc64
        else if (lookup(buffer, "cpu MHz", &cpuinf->cur_freq_MHz));
        else if (lookup(buffer, "processor", &cpu_process))
        {
            cpuinf->cpu_process = atoi(cpu_process) + 1;
            free(cpu_process);
        }
        else if (lookup(buffer, "cpu cores", &corenums))
        {
            cpuinf->corenums = atoi(corenums);
            free(corenums);
        }
    }
    fclose(fp);

    if (cpuinf->flags)
    {
        if (strstr(cpuinf->flags, " svm "))
            cpuinf->virt = strdup("svm");
        else if (strstr(cpuinf->flags, " vmx "))
            cpuinf->virt = strdup("vmx");
    }

#endif
    return;
}

#define GET_VALUE(name, default_val)    do{\
                                        if (! cpuinf)\
                                            _get_cpu_info();\
                                        return cpuinf ? cpuinf->name : default_val;\
                                        }while(0)

const char* kdk_cpu_get_arch()
{
    GET_VALUE(arch, NULL);
}

const char* kdk_cpu_get_vendor()
{
    GET_VALUE(vendor, NULL);
}

const char* kdk_cpu_get_model()
{
    GET_VALUE(model, NULL);
}

const char* kdk_cpu_get_freq_MHz()
{
    GET_VALUE(cur_freq_MHz, NULL);
}

unsigned int kdk_cpu_get_corenums()
{
    GET_VALUE(corenums, 0);
}

const char* kdk_cpu_get_virt()
{
    GET_VALUE(virt, NULL);
}

unsigned int kdk_cpu_get_process()
{
    GET_VALUE(cpu_process, 0);
}